#include "Container.h"
#include "helper.h"
#include "gl_helper.h"
#include <GL/glut.h>
#include <cmath>

Container::Container(float stiffness, float damping, bool computePressureField) :
  _stiffness(stiffness),
  _damping(damping),
  _computePressureField(computePressureField)
  {}


void Container::drawPressureField() const {
  if(_computePressureField) {
    glColor3f(0.0, 0.0, 1.0);
    glBegin(GL_LINES);
      std::vector<std::pair<Vector, Vector> >::const_iterator it = _pressureField.begin();
      while(it != _pressureField.end()) {
        glVertex3f(it->first.x, it->first.y, it->first.z);
        glVertex3f(it->first.x + it->second.x,
                   it->first.y + it->second.y,
                   it->first.z + it->second.z);
        it++;
      }
    glEnd();
  }
}

Vector Container::getResultantPressureField(float radius) const {
  std::vector<std::pair<Vector, Vector> >::const_iterator it = _pressureField.begin();
  Vector resultant;
  while(it != _pressureField.end()) {
    resultant += it->second;
    it++;
  }

  resultant.z = 0; //Don't give a shit about z-component

  return resultant * radius;
}



/********* ContainerBox *******/


ContainerBox::ContainerBox(float stiffness, float damping, 
                           const Vector &bottom, const Vector &top,
                           bool computePressureField) :
  Container(stiffness, damping, computePressureField),
  _bottom(bottom),
  _top(top)
  {}

float ContainerBox::volume() const {
  return ABS(_bottom.x - _top.x) *
         ABS(_bottom.y - _top.y) *
         ABS(_bottom.z - _top.z);
}

Particle ContainerBox::getARandomParticle() const {
  Vector x(RAND(_bottom.x, _top.x),
           RAND(_bottom.y, _top.y),
           RAND(_bottom.z, _top.z));

  return Particle(x);
}

void ContainerBox::resolveCollisions(Particle &p, const Parameters &pm) {
  Vector &x = p._x;
  Vector &ve = p._ve;
  Vector &a = p._a;

  const float ss     = pm.simscale;
  const float radius = pm.particle_radius;

  float diff, adj;
  Vector norm;

  //Z-axis
  diff = 2 * radius - (x.z - _bottom.z) * ss;
  if (diff >= EPS) {
    norm.x = 0;
    norm.y = 0;
    norm.z = 1;

    adj = _stiffness * diff - _damping * (norm * ve);
    a += adj * norm;

    if(_computePressureField && p._p > 0) 
      _pressureField.push_back(std::make_pair(x, -1 * ss * p._p * norm));
  }


  diff = 2 * radius - (_top.z - x.z) * ss;
  if (diff >= EPS) {
    norm.x = 0;
    norm.y = 0;
    norm.z = -1;

    adj = _stiffness * diff - _damping * (norm * ve);
    a += adj * norm;
  
    if(_computePressureField && p._p > 0) 
      _pressureField.push_back(std::make_pair(x, -1 * ss * p._p * norm));
  }

    //X-axis
  diff = 2 * radius - (x.x - _bottom.x) * ss;
  if (diff >= EPS) {
    norm.x = 1;
    norm.y = 0;
    norm.z = 0;

    adj = _stiffness * diff - _damping * (norm * ve);
    a += adj * norm;

    if(_computePressureField && p._p > 0) 
      _pressureField.push_back(std::make_pair(x, -1 * ss * p._p * norm));
  }

  diff = 2 * radius - (_top.x - x.x) * ss;
  if (diff >= EPS) {
    norm.x = -1;
    norm.y = 0;
    norm.z = 0;

    adj = _stiffness * diff - _damping * (norm * ve);
    a += adj * norm;

    if(_computePressureField && p._p > 0) 
      _pressureField.push_back(std::make_pair(x, -1 * ss * p._p * norm));
  }

    //Y-axis
  diff = 2 * radius - (x.y - _bottom.y) * ss;
  if (diff >= EPS) {
    norm.x = 0;
    norm.y = 1;
    norm.z = 0;

    adj = _stiffness * diff - _damping * (norm * ve);
    a += adj * norm;

    if(_computePressureField && p._p > 0) 
      _pressureField.push_back(std::make_pair(x, -1 * ss * p._p * norm));
  }

  diff = 2 * radius - (_top.y - x.y) * ss;
  if (diff >= EPS) {
    norm.x = 0;
    norm.y = -1;
    norm.z = 0;

    adj = _stiffness * diff - _damping * (norm * ve);
    a += adj * norm;

    if(_computePressureField && p._p > 0) 
      _pressureField.push_back(std::make_pair(x, -1 * ss * p._p * norm));
  }
}

void ContainerBox::draw() const {
  const Vector &t = _top,
               &b = _bottom;

  glColor3f(0.0, 0.0, 1.0);
  glBegin(GL_LINES);
    glVertex3f(b.x, b.y, b.z); glVertex3f(t.x, b.y, b.z);
    glVertex3f(b.x, t.y, b.z); glVertex3f(t.x, t.y, b.z);
    glVertex3f(b.x, b.y, b.z); glVertex3f(b.x, t.y, b.z);
    glVertex3f(t.x, b.y, b.z); glVertex3f(t.x, t.y, b.z);

    glVertex3f(b.x, b.y, b.z); glVertex3f(b.x, b.y, t.z);
    glVertex3f(b.x, b.y, t.z); glVertex3f(t.x, b.y, t.z);
    glVertex3f(t.x, b.y, t.z); glVertex3f(t.x, b.y, b.z);

  glEnd();

  drawPressureField();
}

/**** ContainerSphere ***/

ContainerSphere::ContainerSphere(float stiffness, float damping,
                                 const Vector &center, float radius) :
  Container(stiffness, damping),
  _center(center),
  _radius(radius),
  _radius2(radius * radius),
  _pi(3.14159)
  {}

float ContainerSphere::volume() const {
  return 4.0 / 3.0 * _pi * _radius2 * _radius;
}

Particle ContainerSphere::getARandomParticle() const {
  float r = RAND(0.0, _radius),
        theta = RAND(- _pi / 2.0, _pi / 2.0),
        phi = RAND(- _pi, _pi);

  Vector x(r * cos(theta) * cos(phi),
           r * cos(theta) * sin(phi),
           r * sin(theta));

  return Particle(_center + x);
}

void ContainerSphere::resolveCollisions(Particle &p, const Parameters &pm) {
  Vector &x = p._x,
         &ve = p._ve,
         &a = p._a;

  if (F(x) < 0)
    return;

  float diff = ((x - pm.particle_radius) - _center).norm() - _radius;
  Vector norm = (_center - x).normalize();


  diff *= pm.simscale;

  float adj = _stiffness * diff - _damping * (norm * ve);
  a += adj * norm;

  if(_computePressureField && p._p > 0) 
    _pressureField.push_back(std::make_pair(x, -1 * pm.simscale * p._p * norm));

}

float ContainerSphere::F(Vector &x) const {
  return (x - _center).sqrnorm() - _radius2;
}

void ContainerSphere::draw() const {
  glEnable(GL_NORMALIZE);
  glPushMatrix();
    glTranslatef(_center.x, _center.y, _center.z);
    glutWireSphere(_radius, 10, 10);
  glPopMatrix();

}
