/*
 * Copyright (c) 2012 Chabot Simon
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS 0AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef VOLUME_XLOZNK2W
#define VOLUME_XLOZNK2W

#include "Vector.h"
#include "Particle.h"
#include "Parameters.h"

#include <utility>

/**
 * \file Container.h
 * \author Simon Chabot <simon.chabot@etu.utc.fr>
 * \brief Define a container
 */


/**
 * \class Container
 * Pure abstract class, used to describe a container, and which functions are needed.
 **/

class Container {
  public :
    /**
     * \name Container
     * \param stiffness
     * \param damping
     * \param computePressureField
     **/
    Container(float stiffness = 20000.0, float damping = 256.0, bool computePressureField = false);

    /**
     * \fn float volume()
     * \brief Return the volume of the container
     **/
    virtual float volume() const = 0;

    /**
     * \fn Particle getARandomParticle() const = 0;
     * \brief Return a particle, at a random position, inside the container
     **/
    virtual Particle getARandomParticle() const = 0;

    /**
     * \fn void resolveCollisions(Particle &particle, const Parameters &pm)
     * \brief Chech if the particle is inside or outside of the container. Is it's outside, it has
     * to apply some boundary rules
     *
     * \param particle : check for this particle
     * \param pm       : actual simulation parameters
     **/
    virtual void resolveCollisions(Particle &particle, const Parameters &pm) = 0;

    /**
     * \fn void draw()
     * \brief Called to draw the container
     **/
    virtual void draw() const = 0;

    /**
     * \fn resetPressureField()
     * \brief reset the pressure field computation
     **/
    void resetPressureField() { _pressureField.clear(); }

    /**
     * \fn getResultantPressureField()
     * \brief Compute the pressure field resultant (basic integration)
     * \param radius : radius of a fluid particle
     * \return resultant force
     **/
    Vector getResultantPressureField(float radius) const;
    
    void drawPressureField() const;
    void togglePressureFieldComputation() { _computePressureField = !_computePressureField; }
    bool pressureFieldCompution() const { return _computePressureField; }

  protected :
    float _stiffness, _damping;

    std::vector<std::pair<Vector, Vector> > _pressureField;
    bool _computePressureField;

};

class ContainerBox : public Container {
  public:
    ContainerBox (float stiffness, float damping, 
                  const Vector &bottom = Vector(), const Vector &top = Vector(),
                  bool computePressureField = false);

    float volume() const;
    Particle getARandomParticle() const;
    void resolveCollisions(Particle &particle, const Parameters &pm);
    void draw() const;

    const Vector& top() { return _top; }
    const Vector& bottom() { return _bottom; }

  private:

  /*
   *  Exemplification
   *    +---------------- T
   *   /                 /|
   *  /                 / |
   * +-----------------+  |
   * |                 |  +
   * |                 | /
   * B-----------------+/
   *
   */

    Vector _bottom, _top;
};

class ContainerSphere : public Container {
  public :
    ContainerSphere (float stiffness, float damping,
                     const Vector &center = Vector(), float radius = 10);

    float volume() const;
    Particle getARandomParticle() const;
    void resolveCollisions(Particle &particle, const Parameters &pm);
    void draw() const;

  private :
    float F(Vector &x) const; /* F(x) < 0 if x is inside
                                 F(x) > 0 if x is outside
                                 F(x) = 0 if x is on the surface
                               */

    Vector _center;
    float _radius, _radius2;
    
    const float _pi;
};


#endif /* end of include guard: VOLUME_XLOZNK2W */
