#include "Kdtree.h"


KdNode::KdNode(std::vector<std::pair<int, Vector> >::const_iterator const &b, 
       std::vector<std::pair<int, Vector> >::const_iterator const &e):
  leaf(true) {
  leafdata.begin = &(*b);
  leafdata.end = leafdata.begin + (e-b);
}

KdNode::KdNode(float sep, int dim, KdNode *c1, KdNode *c2):
  leaf(false) {
  nodedata.separator = sep;
  nodedata.dimension = dim;
  nodedata.left = c1;
  nodedata.right = c2;
}

KdNode::~KdNode() {
  if (!leaf) {
    delete nodedata.left;
    delete nodedata.right;
  }
}

KdNode* KdTree::splitCell(std::vector<std::pair<int, Vector> >::iterator begin, 
                    std::vector<std::pair<int, Vector> >::iterator end, 
                    Vector const &bmin, Vector const &bmax, int depth) {
  int n = end-begin;

  if (n <= bucket_size)
    return new KdNode(begin, end);

  int dim = depth % 3;
  
  std::vector<std::pair<int, Vector> >::iterator mid = begin + n/2;
  
  // partially sort the range we're operating on
  std::nth_element(begin, mid, end, point_comp(dim));
  
  float separator = mid->second[dim];
  
  Vector divmin = bmin, divmax = bmax;
  divmin[dim] = separator;
  divmax[dim] = separator;
  
  return new KdNode(separator, dim, 
                    splitCell(begin, mid, bmin, divmax, depth + 1), 
                    splitCell(mid, end, divmin, bmax, depth + 1));
}

void KdTree::drawKdNode(KdNode const *node, Vector const &bmin, Vector const &bmax, Vector const &x) const {
  if (node->leaf)
    return;
  
  if (node->nodedata.dimension == 0) {
    glColor4f(0.6, 0.1, 0.1, 0.3);
    glVertex3f(node->nodedata.separator, bmin.y, bmin.z);
    glVertex3f(node->nodedata.separator, bmax.y, bmin.z);
    glVertex3f(node->nodedata.separator, bmax.y, bmax.z);
    glVertex3f(node->nodedata.separator, bmin.y, bmax.z);
  } else if(node->nodedata.dimension == 1) {
    glColor4f(0.1, 0.6, 0.1, 0.3);
    glVertex3f(bmin.x, node->nodedata.separator, bmin.z);
    glVertex3f(bmax.x, node->nodedata.separator, bmin.z);
    glVertex3f(bmax.x, node->nodedata.separator, bmax.z);
    glVertex3f(bmin.x, node->nodedata.separator, bmax.z);
  } else {
    glColor4f(0.1, 0.1, 0.6, 0.3);
    glVertex3f(bmin.x, bmin.y, node->nodedata.separator);
    glVertex3f(bmax.x, bmin.y, node->nodedata.separator);
    glVertex3f(bmax.x, bmax.y, node->nodedata.separator);
    glVertex3f(bmin.x, bmax.y, node->nodedata.separator);
  }
  
  Vector divmin = bmin, divmax = bmax;
  divmin[node->nodedata.dimension] = node->nodedata.separator;
  divmax[node->nodedata.dimension] = node->nodedata.separator;  
  
  drawKdNode(node->nodedata.left, bmin, divmax, x);
  drawKdNode(node->nodedata.right, divmin, bmax, x);
}

// actually create the tree
void KdTree::init() {
  if (positions.empty())
    return;
  
  // compute bounding box
  bbox_min = bbox_max = positions[0].second;

  for (int i = 1; i < positions.size(); ++i) {
    const Vector &p = positions[i].second;

    if (p.x < bbox_min.x) 
      bbox_min.x = p.x;
    else if (p.x > bbox_max.x)
      bbox_max.x = p.x;
      
    if (p.y < bbox_min.y)
      bbox_min.y = p.y;
    else if (p.y > bbox_max.y)
      bbox_max.y = p.y;
    
    if (p.z < bbox_min.z)
      bbox_min.z = p.z;
    else if (p.z > bbox_max.z)
      bbox_max.z = p.z;
  }
  
  root = splitCell(positions.begin(), positions.end(), bbox_min, bbox_max);    
}

void KdTree::neighbors(Vector const &x, std::vector<NeighborData> &nbs) const {
  nbs.clear();
  if (!root)
    return;

  // descend down tree to smallest cell that fully contains query 
  KdNode const *node = root;
  // node has bbox bmin-bmax
  Vector bmin = bbox_min, bmax = bbox_max;
  while (!node->leaf) {
    Vector divmin = bmin, divmax = bmax;
    divmin[node->nodedata.dimension] = node->nodedata.separator;
    divmax[node->nodedata.dimension] = node->nodedata.separator;
    if (containsSphere(x, bmin, divmax)) {
      node = node->nodedata.left;
      bmax = divmax;
    } else if (containsSphere(x, divmin, bmax)) {
      node = node->nodedata.right;
      bmin = divmin;
    } else
      break;
  }

  gatherSamples(node, bmin, bmax, x, nbs);
}

void KdTree::draw(Vector const &p) const {

  glBlendFunc(GL_SRC_ALPHA,GL_ONE);  // Blending Function For Translucency Based On Source Alpha Value
  glEnable(GL_BLEND);
  glBegin(GL_QUADS);
  // draw dividers
    drawKdNode(root, bbox_min, bbox_max, p);
  glEnd();    
  glDisable(GL_BLEND);
}
