#ifndef KDTREE_HH
#define KDTREE_HH

// for drawing routine
#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

#include <vector>
#include <algorithm>
#include <limits>
#include "Vector.h"
#include "Neighbordata.h"

class KdNode {
public:
  bool leaf;
  union {
    struct {
      KdNode *left;
      KdNode *right;
      float separator;
      int dimension;
    } nodedata;
    struct {
      std::pair<int, Vector> const *begin;
      std::pair<int, Vector> const *end;
    } leafdata;
  };
  
  KdNode(std::vector<std::pair<int, Vector> >::const_iterator const &b, 
         std::vector<std::pair<int, Vector> >::const_iterator const &e);
  KdNode(float sep, int dim, KdNode *c1, KdNode *c2);
  ~KdNode();
};

struct point_comp {
  int dim;
  point_comp(int d): dim(d) {};
  
  inline bool operator() (std::pair<int, Vector> const &x, std::pair<int, Vector> const &y) const {
    return x.second[dim] < y.second[dim];
  }
};

class KdTree {
  float query_radius, query_radius2;
  int bucket_size;
  Vector bbox_min, bbox_max;
  
  // storage for points
  std::vector<std::pair<int, Vector> > positions;
  
  // index tree
  KdNode *root;
  
  KdNode *splitCell(std::vector<std::pair<int, Vector> >::iterator begin, 
                    std::vector<std::pair<int, Vector> >::iterator end, 
                    Vector const &bmin, Vector const &bmax, int depth = 0);

  inline bool containsSphere(Vector const &x, Vector const &bmin, Vector const &bmax) const {
    float minx = x.x - query_radius;
    float maxx = x.x + query_radius;
    float miny = x.y - query_radius;
    float maxy = x.y + query_radius;
    float minz = x.z - query_radius;
    float maxz = x.z + query_radius;
    
    return 
      minx >= bmin.x && maxx < bmax.x &&
      miny >= bmin.y && maxy < bmax.y &&
      minz >= bmin.z && maxz < bmax.z;
  }

  inline bool intersectsSphere(Vector const &x, Vector const &bmin, Vector const &bmax) const {
    float dist = 0;    
    for(int i = 0; i < 3; ++i) {
      if(x[i] < bmin[i]) {
        float d = bmin[i] - x[i];
        dist += d * d;
      } else if(x[i] > bmax[i]) {
        float d = x[i] - bmax[i];
        dist += d * d;
      }
    }
    
    return dist <= query_radius2;
  }
  
  inline void gatherSamples(KdNode const *node, Vector const &bmin, Vector const &bmax, Vector const &x, 
                            std::vector<NeighborData> &nbs) const {
    if (node->leaf) {
      // check all in this bucket and add those within range
      for (std::pair<int, Vector> const * it = node->leafdata.begin; it != node->leafdata.end; ++it) {
        Vector dir = it->second - x;
        float d_squared = dir.sqrnorm();
        if (d_squared <= query_radius2) {
          NeighborData data;
          data.idx = it->first;
          data.d_squared = d_squared;
          data.d = sqrt(d_squared);
          if (data.d != 0) 
            data.d_normalized = dir/data.d;
          else
            data.d_normalized = Vector(0,0,0);
          nbs.push_back(data);
        }        
      }
    } else {
      // check if children intersect
      Vector divmin = bmin, divmax = bmax;
      divmin[node->nodedata.dimension] = node->nodedata.separator;
      divmax[node->nodedata.dimension] = node->nodedata.separator;
      if (intersectsSphere(x, bmin, divmax))
        gatherSamples(node->nodedata.left, bmin, divmax, x, nbs);
      if (intersectsSphere(x, divmin, bmax))
        gatherSamples(node->nodedata.right, divmin, bmax, x, nbs);
    }
  }
  
  void drawKdNode(KdNode const *node, Vector const &bmin, Vector const &bmax, Vector const &x) const;
  
public:
  KdTree(int bucketsize = 30): bucket_size(bucketsize), root(NULL) {
  }
  
  ~KdTree() {
    delete root;
  }
  
  inline KdNode const *getRoot() const {
    return root;
  }
  
  inline void getBBox(Vector &bmin, Vector &bmax) const {
    bmin = bbox_min;
    bmax = bbox_max;
  }

  inline float queryRadius() const {
    return query_radius;
  }
  
  // sets the query radius. This is cheap, and can be done before each query
  inline float queryRadius(float r) {
    query_radius = r;
    query_radius2 = r*r;
  }
    
  inline void clear() {
    if (root)
      delete root;
    root = NULL;
    positions.clear();
    bbox_min = Vector(0,0,0);
    bbox_max = Vector(-1,-1,-1);
  }

  inline void insert(int idx, Vector const &x) {
    positions.push_back(std::make_pair(idx, x));
  }
  
  // actually create the tree
  void init();
  void neighbors(Vector const &x, std::vector<NeighborData> &nbs) const;
  void draw(Vector const &p = Vector(std::numeric_limits<double>::infinity(), 
                                     std::numeric_limits<double>::infinity(), 
                                     std::numeric_limits<double>::infinity())) const;
};

#endif

