/*
 * Copyright (c) 2012 Martin Wicke
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS 0AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "Kernel.h"

std_kernel::std_kernel(float h_) :
  h(h_),
  h_squared(h*h),
  h_third(h_squared*h),
  h_fourth(h_squared*h_squared) 
  {}
  
float std_kernel::value(float d, float d_squared) const {
  if (d_squared >= h_squared)
    return 0.f;
  else {
    float x = 1.f - d_squared/h_squared;
    return 315.0f/(64.0f * M_PI*h_third) * x * x * x;
  }

}
  
float std_kernel::first_derivative(float d, float d_squared) const {
  if (d >= h)
    return 0.f;
  else {
    float x = 1.f - d_squared/h_squared;
    return -945.0f/ (32.0f * M_PI * h_third) * (d / h_squared) * x * x;
  }
}

Vector std_kernel::gradient(NeighborData const &nd) const {
  return - first_derivative(nd.d, nd.d_squared) * nd.d_normalized;
}

float std_kernel::second_derivative(float d, float d_squared) const {
  if (d_squared >= h_squared)
    return 0.f;
  else {
    float x = d_squared/h_squared;
    return (945.0f / (32.0f * M_PI * h_fourth)) * (1 - x) * (5 * x - 1);
  }
}

spiky_kernel::spiky_kernel(float h_): 
  h(h_),
  h_squared(h*h), 
  h_third(h_squared*h),
  h_fourth(h_squared*h_squared) 
  {}

float spiky_kernel::value(float d, float d_squared) const {
  if (d >= h)
    return 0.f;
  else {
    float x = 1.f - d/h;
    return 15.f / (M_PI*h_third) * x * x * x;
  }
}

float spiky_kernel::first_derivative(float d, float d_squared) const {
  if (d >= h)
    return 0.f;
  else {
    float x = 1.f - d/h;
    return -45.f / (M_PI*h_fourth) * x * x;
  }
}

Vector spiky_kernel::gradient(NeighborData const &nd) const {
  return - first_derivative(nd.d, nd.d_squared) * nd.d_normalized;
}

float spiky_kernel::second_derivative(float d, float d_squared) const {
  if (d >= h)
    return 0.f;
  else {
    float x = 1.f - d/h;
    return 90.f / (M_PI*h_third * h_squared) * x;
  }
}
