#include "Sph.h"
#include <fstream>

#include <GL/glut.h>
#include <time.h>  /* Needed for the random function */
#include <cstdlib> /* Idem */

#include "gl_helper.h"
#include "helper.h"

SPH::SPH(const Parameters &pm, const std::string &loadingFile) {

  setParameters(pm);
  _N = 0;

  distributeParticles(loadingFile);

  _neighbors.resize(_params.N);
  for(int i = 0 ; i < _params.N ; i++)
    _neighbors[i].reserve(_params.numberOfNeighbors);
}

void SPH::setParameters(const Parameters &pm) {
  _params = pm;

  compute_parameters();

  _vKernel = viscosity_kernel(_h);
  _pKernel = pressure_kernel(_h);
  _tree.queryRadius(_h / _params.simscale);
}

void SPH::compute_parameters() {
  /*
   * h = [(3 * volume * x) / (4 * pi * total number of particles)]^(1/3)
   * m = [(density) * (volume)] / (total number of particles)
   * r = s * (3 * m / (4pi rho))^(1/3)
   */

  
  float v = _params.fluid_init->volume()* pow(_params.simscale, 3);
  float s = 0.6; //A user constante to act on particles' radius

  _m = (_params.rho0 * v)/_params.N;
  _params.particle_radius = s * pow(3.0*_m/(4.0*M_PI*_params.rho0), 1.0/3.0);
  _h = pow((3.0 * v * _params.numberOfNeighbors) / (4.0 * M_PI * _params.N), 1.0/3.0);

  std::cerr << "v = " << v << "m^3" << std::endl;
  std::cerr << "rho0 = " << _params.rho0 << "kg/m^3" << std::endl;
  std::cerr << "m = " << _m << "kg" << std::endl;
  std::cerr << "r = " << _params.particle_radius << "m" << std::endl;
  std::cerr << "h = " << _h << "m" << std::endl;
}

  
void SPH::distributeParticles(const std::string &loadingFile) {

  _particles.clear();
  _particles.reserve(_params.N);

  if (loadingFile.empty()) {
    #pragma omp for
    for (int i = 0; i <  _params.N ; i++) {
      _particles.push_back(_params.fluid_init->getARandomParticle());
      _N++;
    }
  } else {
    std::ifstream data_file(loadingFile.c_str());
    std::string line;
    Vector x;
    while(getline(data_file, line)) {
      x << line;
      _particles.push_back(Particle(x));
      _N++;
    }
  }
}

void SPH::drawParticles(float *viewmat) const {
  glEnable(GL_NORMALIZE);
  glLoadMatrixf(viewmat);

  float r = _params.particle_radius / _params.simscale;

  for(int i = 0 ; i < _N ; i++) {
    const Vector &p = _particles[i]._x;
    unsigned int clr = _particles[i]._clr;

    glPushMatrix();
      glTranslatef(p.x, p.y, p.z);
      glScalef(r, r, r);
      glColor4f(RED(clr), GRN(clr), BLUE(clr), ALPH(clr));
      drawSphere();
    glPopMatrix();

  }
}

void SPH::updateTree() {
  _tree.clear();
  for(int i = 0 ; i < _N ; i++)
    _tree.insert(i, _particles[i]._x);

  _tree.init();
}


void SPH::step() {
  _params.pressureResultant = _params.domain->getResultantPressureField(_params.particle_radius);
  _params.domain->resetPressureField();

  if(_N < _params.N) {
    _params.emitter->emit(_N, _particles, _params.N);
  }

  /* 1st step : update the KdTree */
  updateTree();

  /* 2nd step : compute neighbors, compute densities and pressures */
  #pragma omp for
  for (int i = 0 ; i < _N ; i++) {
    _tree.neighbors(_particles[i]._x, _neighbors[i]);
    _particles[i]._rho = 0;

    for (int j = 0 ; j < _neighbors[i].size() ; j++) {
      _neighbors[i][j].d *= _params.simscale;     //distance in m
      _neighbors[i][j].d_squared = _params.simscale * _params.simscale;

      _particles[i]._rho += _m * _pKernel.value(_neighbors[i][j].d, _neighbors[i][j].d_squared);
    }

    _particles[i]._p = (pow(_params.intstiff, 2) * _params.rho0 / 7.0) * (pow(_particles[i]._rho / _params.rho0, 7) - 1);
    if(_particles[i]._p < 0)
      _particles[i]._p *= _params.pressureCorrection;
  }

  /* 3rd step : compute forces, accelerations and velocities */
  #pragma omp for
  float max_speed = 0;
  for (int i = 0 ; i < _N ; i++) {
    Particle &pi = _particles[i];
    Vector f;

    for (int j = 0 ; j < _neighbors[i].size() ; j++) {
      Particle &pj = _particles[_neighbors[i][j].idx];
      float &d = _neighbors[i][j].d;
      float &d_squared = _neighbors[i][j].d_squared;

      // Pressure force
      f -= _m * pi._rho * (pj._p / (pj._rho * pj._rho) + pi._p / (pi._rho * pi._rho)) * _pKernel.gradient(_neighbors[i][j]);

      // Viscosity force 
      f += _params.visc / pi._rho * (_m / (pj._rho * pj._rho)) * (pj._ve - pi._ve) * _vKernel.second_derivative(d, d_squared);

    }

    if(ABS(pi._ve.x) > max_speed)
      max_speed = ABS(pi._ve.x);

    pi._a = f / pi._rho + _params.gravity - (_params.acc + _params.feedback + _params.pressureResultant);
  }

  /* 4th step : move particles according to acceleration */
  Vector vnext;

  #pragma omp for
  for (int i = 0 ; i < _N ; i++) {
    Particle &pi = _particles[i];
    Vector &a = pi._a;
    Vector &x = pi._x;
    Vector &ve = pi._ve;
    Vector &v = pi._v;
    
    //Velocity limiting
    if (a.sqrnorm() > _params.maxvel * _params.maxvel) 
      a *= _params.maxvel / a.norm();


    //Boundary conditions
    _params.domain->resolveCollisions(pi, _params);


    /*****  Leapfrog integration **
     *
     * v(t + 1/2) = v(t - 1/2) + a*dt
     * x(t + 1)   = x(t) + v(t + 1/2) * dt
     *
     * v(t + 1) = 1/2 * (v(t + 1/2) + v(t - 1/2)) //Needed for acceleration computation
     **/

    vnext = a * _params.dt + v;
    ve = (v + vnext) * 0.5;
    v = vnext;
    x += v * _params.dt / _params.simscale;
    
    float gamma = pi._p / 2500;
    float beta = 1;

    if (gamma > 1) gamma = 1;
    if (gamma <= 0) gamma = 0;
    if (gamma < -1) gamma = -1;

/* Pressure */
    float red = (gamma > 0.5) ?  2*gamma - 1 : 0;
    float blue = (gamma <= 0.5) ?  1 - 2*gamma : 0;
    float green = (gamma < 0.5) ? gamma : 1-gamma;
    float alpha = (1.0 - beta) * 0.5 + beta * 1;

/* Speed */
/*
    float red = (gamma > 0) ? 0 : gamma;
    float blue = (gamma < 0) ? 0 : gamma;
    float green = (1 - ABS(gamma)) * 0.5;
    float alpha = beta;
*/
    pi._clr = COLORA(red, green, blue, alpha);
  }
}

void SPH::save(const std::string &filename) const {
  std::ofstream myFile(filename.c_str());

  if(myFile) {
    std::vector<Particle>::const_iterator it = _particles.begin();
    while(it != _particles.end()) {
      myFile << it->_x << std::endl;
      it++;
    }
    std::cout << "Saved" << std::endl;
  }
}

  

