/*
 * Copyright (c) 2012 Martin Wicke, Simon Chabot
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS 0AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "Parameters.h"
#include "Vector.h"
#include "Particle.h"
#include "Neighbordata.h"
#include "Kernel.h"
#include "Kdtree.h"
#include "Container.h"
#include "Emitter.h"
#include <vector>

#ifndef SPH_PIL4JM5P
#define SPH_PIL4JM5P

/**
 * \file Sph.h
 * \author Martin Wicke <wicke@eecs.berkeley.edu>
 * \author Simon Chabot <simon.chabot@etu.utc.fr>
 */

/**
 * \class SPH
 */

class SPH {
  public:
    SPH (const Parameters &pm, const std::string &loadingFile = "");

    Parameters& getParams() { return _params ; }
    void setParameters(const Parameters &pm);
    void compute_parameters(); 
    void drawParticles(float *viewmat) const;
    void step();
    const std::vector<Particle>& getParticles() { return _particles; }
    void save(const std::string &filename = "save.tx00") const;
    int getNb() const { return _N; };

    const KdTree& tree() { return _tree; }
  
  private:
    typedef spiky_kernel pressure_kernel;
    typedef std_kernel   viscosity_kernel;

    void distributeParticles(const std::string &loadingFile = ""); /*! Compute the initial positions */
    void updateTree(); /*! Destroy and rebuild the tree */
     

    float _h;    /*! Smoothing radius (computed from spacing) */
    float _m;    /*! Mass of each particle (computed from spacing and density) */
    int _N;      /*! Current number of particles */
    Parameters _params; /*! Simultation's parameters */
    std::vector<Particle> _particles;  /*! Particle storage */
    KdTree _tree;   /*! Spacial search data structure                             */
    std::vector<std::vector<NeighborData> > _neighbors;  /*! A record of the neighbors of a given particle */
    pressure_kernel _pKernel;  /*! Kernel used for pressure computation  */
    viscosity_kernel _vKernel; /*! Kernel used for viscosity computation */

};


#endif /* end of include guard: SPH_PIL4JM5P */
