/*
 * Copyright (c) 2012 Martin Wicke, Simon Chabot
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS 0AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#ifndef VECTOR_HH
#define VECTOR_HH

#include <iostream>
#include <math.h>
#include <cstdlib>

/**
 * \file Vector.h
 * \author Martin Wicke <wicke@eecs.berkeley.edu>
 * \author Simon Chabot <simon.chabot@etu.utc.fr>
 * \brief Describes a 3D vector
 */

struct Vector
{
  double x, y, z;
  
  Vector(double _x = 0, double _y = 0, double _z = 0) : x(_x), y(_y), z(_z) {};
  
  /**
   * \fn inline double const &operator[](unsigned int i) const
   * \brief Get the access to the i-th component
   */
  inline double const &operator[](unsigned int i) const {
    return *(&x + i);
  }
  
  inline double &operator[](unsigned int i) {
    return *(&x + i);
  }
  
  inline Vector operator-(Vector const &v) const {
    return Vector(x-v.x, y-v.y, z-v.z);
  }
  
  inline Vector operator+(Vector const &v) const {
    return Vector(x+v.x, y+v.y, z+v.z);
  }
  
  inline Vector operator*(double s) const {
    return Vector(s*x, s*y, s*z);
  }
  
  inline Vector operator/(double s) const {
    return operator*(1/s);
  }
  
  inline Vector &operator+=(Vector const &v) {
    x += v.x;
    y += v.y;
    z += v.z;
    return *this;
  }
  
  inline Vector &operator-=(Vector const &v) {
    x -= v.x;
    y -= v.y;
    z -= v.z;
    return *this;
  }
  
  inline Vector &operator*=(double d) {
    x *= d;
    y *= d;
    z *= d;
    return *this;
  }
  
  inline Vector &operator/=(double d) {
    return *this *= 1.0/d;
  }

  inline Vector &operator=(Vector const &v) {
    x = v.x;
    y = v.y;
    z = v.z;
    return *this;
  }

  inline bool operator==(Vector const &v) const {
    return (v.x == x) && (v.y == y) && (v.z == z);
  }

  inline bool operator!=(Vector const &v) const {
    return !(*this == v);
  }
  
  /**
   * \fn inline double operator*(Vector const &v) const
   * \brief Dot product
   */
  inline double operator*(Vector const &v) const {
    return v.x*x + v.y*y + v.z*z;
  }
  
  /**
   * \fn inline Vector operator^(Vector const &v) const
   * \brief Cross product
   */
  inline Vector operator^(Vector const &v) const {
    return Vector(y*v.z - z*v.y, z*v.x - x*v.z, x*v.y - y*v.x);
  }
  
  /**
   * \fn inline double sqrnorm() const
   * \brief Compute the squared norm
   */
  inline double sqrnorm() const {
    return *this * *this;
  }
  
  /**
   * \fn inline double norm() const
   * \brief Compute the norm
   */
  inline double norm() const {
    return sqrt(sqrnorm());
  }
  
  inline Vector &normalize() {
    double n = this->norm();
    if (n != 0)
      *this *= (1.0/n);
    
    return *this;
  }

  /**
   * \fn inline Vector rotated(float alpha, Vector v = Vector(0, 0, 1)) const
   * \brief Give the rotated image of the vector
   * \param alpha angle to rotate
   * \param v with respect to this vector (must be e_x, e_y or e_z !)
   */
  inline Vector rotated(float alpha, Vector v = Vector(0, 0, 1)) const {
    Vector r;

    if(v.z == 1) {
      r.x = cos(alpha)*x - sin(alpha)*y;
      r.y = sin(alpha)*x + cos(alpha)*y;
      r.z = z;
    } else if(v.x == 1) {
      r.x = x;
      r.y = cos(alpha)*y - sin(alpha)*z;
      r.z = sin(alpha)*y + cos(alpha)*z;
    } else { //with respect to y-axis
      r.x = cos(alpha)*x + sin(alpha)*z;
      r.y = y;
      r.z = -sin(alpha)*x + cos(alpha)*z;
    }


    return r;
  }

  /**
   * \fn inline Vector &rotate(float alpha, Vector v = Vector(0, 0, 1))
   * \brief Rotate the vector
   * \param alpha angle to rotate
   * \param v with respect to this vector (must be e_x, e_y or e_z !)
   */
  inline Vector &rotate(float alpha, Vector v = Vector(0, 0, 1)) {
    *this = rotated(alpha, v);
    return *this;
  }

  /**
   * \fn inline Vector &reset()
   * \brief Set all components to 0
   **/
  inline Vector &reset() {
    x = y = z = 0;
    return *this;
  }

  inline Vector &operator << (std::string &os) {
    size_t pos_begin = 1, pos_end = os.find(", ");

    x = atof(os.substr(pos_begin, pos_end - pos_begin).c_str());
    
    pos_begin = pos_end + 1 ; pos_end = os.find(", ", pos_begin);

    y = atof(os.substr(pos_begin, pos_end - pos_begin).c_str());

    pos_begin = pos_end + 1 ; pos_end = os.find("]", pos_begin);

    z = atof(os.substr(pos_begin, pos_end - pos_begin).c_str());

    return *this;
  }
};

inline Vector operator*(double d, Vector const &v) {
  return v * d;
}

/**
 * \fn inline std::ostream &operator<<(std::ostream &os, Vector const &v)
 * \brief Display a vector
 * \param os stream
 * \param v vector to display
 **/
inline std::ostream &operator<<(std::ostream &os, Vector const &v) {
  return os << '[' << v.x << ", " << v.y << ", " << v.z << ']';
}

#endif
